from datetime import datetime
from typing import List, Optional
from ninja import Schema


class MoonPermisions(Schema):
    view_public_extractions: bool = False
    view_public_observations: bool = False
    view_rentals: bool = False
    edit_rentals: bool = False


class IdName(Schema):
    id: int
    name: str


class Character(Schema):
    character_name: str
    character_id: int
    corporation_name: str
    corporation_id: int
    alliance_name: Optional[str]
    alliance_id: Optional[int]


class Corporation(Schema):
    corporation_name: str
    corporation_id: int
    alliance_name: Optional[str]
    alliance_id: Optional[int]


class OreVolume(Schema):
    type: IdName
    volume: int


class ExtractionEvent(Schema):
    ObserverName: str
    moon: IdName
    extraction_end: datetime
    mined_ore: Optional[List[OreVolume]]


class MoonRenatal(Schema):
    moon: IdName
    system: IdName
    contact: Character
    corporation: Corporation
    price: int
    start_date: datetime


class NewMoonRenatal(Schema):
    moon_id: int
    contact_id: int
    corporation_id: int
    price: int



from django.apps import apps
from django.conf import settings


def get_pub():
    out = getattr(settings, "PUBLIC_MOON_CORPS", [98609787])
    if isinstance(out, list):
        return out
    elif isinstance(out, int):
        return [out]
    elif isinstance(out, str):
        return [int(out)]


PUBLIC_MOON_CORPS = get_pub()  # trust


def discord_bot_active():
    if apps.is_installed("aadiscordbot"):
        import aadiscordbot as ab
        version = ab.__version__.split(".")
        if int(version[0]) >= 3:
            return True
    return False


def get_rental_discord_channel():
    out = getattr(settings, "MOON_RENTAL_CHANNEL", 373640301375651844)
    if isinstance(out, int):
        return out
    elif isinstance(out, str):
        return int(out)



try:
    from django_redis import get_redis_connection
    _client = get_redis_connection("default")
except (NotImplementedError, ModuleNotFoundError):
    from django.core.cache import caches
    default_cache = caches['default']
    _client = default_cache.get_master_client()

cache_client = _client



# Generated by Django 3.2.5 on 2021-08-27 11:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0058_auto_20210808_0635'),
        ('moons', '0009_alter_miningtax_flat_tax_rate'),
    ]

    operations = [
        migrations.AddField(
            model_name='miningobservation',
            name='moon',
            field=models.ForeignKey(blank=True, default=None, null=True,
                                    on_delete=django.db.models.deletion.SET_NULL, to='corptools.mapsystemmoon'),
        ),
    ]



from allianceauth.eveonline.models import EveCharacter
from datetime import timedelta

from typing import List
from django.utils import timezone
from django.utils.timezone import activate

from ninja import NinjaAPI, Form, main
from ninja.security import django_auth
from ninja.responses import codes_4xx

from django.core.exceptions import PermissionDenied
from django.db.models import F, Sum, Q
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from django.conf import settings

from . import models
from corptools.models import MapSystemMoon
from . import schema

from django.db.models import Sum

import logging

logger = logging.getLogger(__name__)


api = NinjaAPI(title="MoonTool API", version="0.0.1",
               urls_namespace='moons:api', auth=django_auth, csrf=True,
               openapi_url=settings.DEBUG and "/openapi.json" or "")


@api.get(
    "/user/permisions",
    response={200: schema.MoonPermisions},
    tags=["User"]
)
def get_user_permisions(request, search_text: str):
    return []


@api.get(
    "/moon/search",
    response={200: List[schema.IdName]},
    tags=["Moons"]
)
def get_moon_search(request, search_text: str):
    return MapSystemMoon.objects.filter(name__icontains=search_text).values("name", id=F("moon_id"))


@api.get(
    "/character/search",
    response={200: List[schema.Character]},
    tags=["Character"]
)
def get_character_search(request, search_text: str):
    return EveCharacter.objects.filter(character_name__icontains=search_text)


@api.get(
    "/extraction/active",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_observer_usage(request, observer_id: int):
    if not request.user.has_perm("moons.access_moons"):
        return []
    time_from = timezone.now() - timedelta(days=6)
    observations = models.MiningObservation.objects \
        .filter(observing_id=observer_id, last_updated__gte=time_from) \
        .values('type_id') \
        .annotate(mined=(Sum('quantity') * F('type_name__volume'))) \
        .annotate(name=F('type_name__name'))

    return observations


@api.get(
    "/extraction/remaining",
    response={200: List[schema.OreVolume]},
    tags=["Observers"]
)
def get_observer_usage(request, observer_id: int):
    if not request.user.has_perm("moons.access_moons"):
        return []
    time_from = timezone.now() - timedelta(days=6)
    observations = models.MiningObservation.objects \
        .filter(observing_id=observer_id, last_updated__gte=time_from) \
        .values('type_id') \
        .annotate(mined=(Sum('quantity') * F('type_name__volume'))) \
        .annotate(name=F('type_name__name'))

    return observations


logger = logging.getLogger(__name__)


api = NinjaAPI(title="MoonTool API", version="0.0.1",
               urls_namespace='moons:api', auth=django_auth, csrf=True,
               openapi_url=settings.DEBUG and "/openapi.json" or "")


@api.get(
    "/user/permisions",
    response={200: schema.MoonPermisions},
    tags=["User"]
)
def get_user_permisions(request, search_text: str):
    return []


@api.get(
    "/moons/search",
    response={200: List[schema.IdName]},
    tags=["Search"]
)
def get_moon_search(request, search_text: str, limit: int = 10):
    return MapSystemMoon.objects.filter(name__icontains=search_text).values("name", id=F("moon_id"))[:limit]


@api.get(
    "/corporations/search",
    response={200: List[schema.Corporation]},
    tags=["Search"]
)
def get_corporation_search(request, search_text: str, limit: int = 10):
    return EveCorporationInfo.objects.filter(corporation_name__icontains=search_text)[:limit]


@api.get(
    "/characters/search",
    response={200: List[schema.Character]},
    tags=["Search"]
)
def get_character_search(request, search_text: str, limit: int = 10):
    return EveCharacter.objects.filter(character_name__icontains=search_text)[:limit]


@api.get(
    "/extraction/active",
    response={200: List[schema.ExtractionEvent]},
    tags=["Observers"]
)
def get_observer_usage(request, observer_id: int):
    if not request.user.has_perm("moons.access_moons"):
        return []
    time_from = timezone.now() - timedelta(days=6)
    observations = models.MiningObservation.objects \
        .filter(observing_id=observer_id, last_updated__gte=time_from) \
        .values('type_id') \
        .annotate(mined=(Sum('quantity') * F('type_name__volume'))) \
        .annotate(name=F('type_name__name'))

    return observations


@api.get(
    "/rental/list",
    response={200: List[schema.MoonRenatal]},
    tags=["Rentals"]
)
def get_moon_rentals(request):
    if not request.user.has_perm("moons.access_moons"):
        return []

    rentals = models.MoonRental.objects.filter(end_date__isnull=True).select_related(
        "moon", "moon__system", "contact", "corporation")
    out = []
    for r in rentals:
        out.append(
            {"moon": {
                "id": r.moon.moon_id,
                "name": r.moon.name
            },
                "system": {
                "id": r.moon.system.system_id,
                "name": r.moon.system.name
            },
                "contact": r.contact,
                "corporation": r.corporation,
                "price": r.price,
                "start_date": r.start_date
            }
        )

    return out


@api.post(
    "/rental/new",
    response={200: schema.MoonRenatal, 403: str},
    tags=["Rentals"]
)
def post_moon_rental_new(request, rental: schema.NewMoonRenatal = Form(...)):
    if models.MoonRental.objects.filter(moon_id=rental.moon_id, end_date__isnull=True).exists():
        return 403, "Moon Already Rented!"
    try:
        char = EveCharacter.objects.get(character_id=rental.contact_id)
    except EveCharacter.DoesNotExist:
        return 403, "Character Unknown to Auth"
    try:
        corp = EveCorporationInfo.objects.get(
            corporation_id=rental.corporation_id)
    except EveCorporationInfo.DoesNotExist:
        return 403, "Corporation Unknown to Auth"
    new_rental = models.MoonRental.objects.create(
        moon_id=rental.moon_id,
        contact=char,
        corporation=corp,
        price=rental.price,
        start_date=timezone.now()
    )
    return 200, {"moon": {
        "id": new_rental.moon.moon_id,
        "name": new_rental.moon.name
    },
        "system": {
        "id": new_rental.moon.system.system_id,
        "name": new_rental.moon.system.name
    },
        "contact": new_rental.contact,
        "corporation": new_rental.corporation,
        "price": new_rental.price,
        "start_date": new_rental.start_date
    }



# Cog Stuff
from typing import Optional
from corptools.models import CorporationAudit, Structure
from discord.embeds import Embed
from allianceauth.services.modules.discord.models import DiscordUser
from discord.ext import commands
from discord.commands import SlashCommandGroup

# AA Contexts
import pprint
from django.conf import settings
from django.utils import timezone
from moons.models import InvoiceRecord, MoonFrack

from moons import app_settings

import logging

logger = logging.getLogger(__name__)


class MoonsCog(commands.Cog):
    """
    All about Moons!
    """

    def __init__(self, bot):
        self.bot = bot

    pinger_commands = SlashCommandGroup(
        "moons", "Moon Module Commands", guild_ids=[int(settings.DISCORD_GUILD_ID)])

    def sender_has_moon_perm(self, ctx):
        id = ctx.author.id
        try:
            has_perm = DiscordUser.objects.get(
                uid=id).user.has_perm("moons.view_all")
            if has_perm:
                return True
            else:
                return False
        except Exception as e:
            return False

    def sender_has_corp_moon_perm(self, ctx):
        id = ctx.author.id
        try:
            has_perm = DiscordUser.objects.get(
                uid=id).user.has_perm("moons.view_corp")
            if has_perm:
                return True
            else:
                return False
        except Exception as e:
            return False

    @pinger_commands.command(name='print_stats', guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def info_slash(self, ctx):
        """
        Print the Uninvocied Mining Stats!
        """

        if not self.sender_has_moon_perm(ctx):
            return await ctx.respond(f"You do not have permision to use this command.", ephemeral=True)

        await ctx.respond(f"Calculating the Moon Stats.")

        last_date = InvoiceRecord.get_last_invoice_date()
        date_str = last_date.strftime('%Y/%m/%d')
        e = Embed(title=f"Last Invoice {date_str}")
        accounts_seen = 0
        locations = set()
        data = InvoiceRecord.generate_invoice_data()
        total_mined = 0
        total_taxed = 0
        # run known people
        for u, d in data['knowns'].items():
            try:
                total_mined += d['total_value']
                total_taxed += d['tax_value']

                accounts_seen += 1
                for l in d['locations']:
                    locations.add(l)
            except KeyError:
                pass  # probably wanna ping admin about it.

        for u, d in data['unknowns'].items():
            try:
                total_mined += d['totals_isk']
                total_taxed += d['tax_isk']
                for l in d['seen_at']:
                    locations.add(l)
            except KeyError:
                pass  # probably wanna ping admin about it.

        e.add_field(name="Known Members",
                    value=f"{accounts_seen}", inline=False)
        e.add_field(name="Unknown Characters",
                    value=f"{len(data['unknowns'])}", inline=False)
        e.add_field(name="Total Mined",
                    value=f"${total_mined:,}", inline=False)
        e.add_field(name="Total Tax", value=f"${total_taxed:,}", inline=False)
        locations = "\n ".join(list(locations))
        e.description = f'Locations tracked so far ({len(list(locations))})\n\n {locations}'

        await ctx.channel.send(embed=e)

    @pinger_commands.command(name='inactive', guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def inactive_moons(self, ctx, own_corp: Optional[bool] = False):
        """
        Print inactive Moons!
        """
        if own_corp:
            if not self.sender_has_corp_moon_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)
            user = DiscordUser.objects.get(
                uid=ctx.author.id).user.profile.main_character
            corps = CorporationAudit.objects.filter(
                corporation__corporation_id=user.corporation_id)
            corp_names = [f"{c.corporation.corporation_name}" for c in corps]
            if corps.count() > 0:
                await ctx.respond(f"Printing Inactive Drills for {', '.join(corp_names)}", ephemeral=True)
                await ctx.author.send(f"Printing Inactive Drills for {', '.join(corp_names)}")
            else:
                await ctx.respond(f"Your corp is not setup in Audit, please contact an admin.", ephemeral=True)

        else:
            if not self.sender_has_moon_perm(ctx):
                return await ctx.respond(f"You do not have permission to use this command.", ephemeral=True)
            corps = CorporationAudit.objects.filter(
                corporation__corporation_id__in=app_settings.PUBLIC_MOON_CORPS)
            corp_names = [f"{c.corporation.corporation_name}" for c in corps]
            if corps.count() > 0:
                await ctx.respond(f"Printing Inactive Drills for {', '.join(corp_names)}")
            else:
                await ctx.respond(f"Public corp is not setup, please contact an admin.")

        tzactive = timezone.now()
        fracks = MoonFrack.objects.filter(
            arrival_time__gte=tzactive, corporation__in=corps).values_list('structure_id')
        structures = Structure.objects.filter(structureservice__name__in=[
            "Moon Drilling"], corporation__in=corps).exclude(structure_id__in=fracks)
        messages = [f"{s.name}" for s in structures]
        n = 10
        chunks = [list(messages[i * n:(i + 1) * n])
                  for i in range((len(messages) + n - 1) // n)]
        for chunk in chunks:
            message = "\n".join(chunk)
            if own_corp:
                await ctx.author.send(f"```{message}```")
            else:
                await ctx.send(f"```{message}```")


def setup(bot):
    bot.add_cog(MoonsCog(bot))
